/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *
 *  Modifications :
 *  Creation P.Delrieu
 *
 */
package org.openmobileis.bundle.update.terminal;

import java.io.File;
import java.io.IOException;

import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.exception.ServiceException;
import org.openmobileis.common.util.exception.SynchroException;
import org.openmobileis.common.util.file.FileUtilities;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.embedded.util.SystemAPI;
import org.openmobileis.services.common.ServiceManager;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.DefaultOpenMSPSyncListener;
import org.openmobileis.synchro.openmsp.client.OpenMSPSynchroManager;
import org.openmobileis.synchro.openmsp.protocol.Command;
import org.openmobileis.synchro.openmsp.protocol.ContainerMessage;
import org.openmobileis.synchro.openmsp.protocol.DataItem;
import org.openmobileis.synchro.openmsp.protocol.Element;
import org.openmobileis.synchro.openmsp.protocol.Item;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Synchro listener use to synchronize a jar file from the server.
 * the jar is unzipped from the user.dir root directory and erase all existing file.
 * The jar is stored in the user.dir/temp directory and named APPLICATION_JAR_FILENAME="fullapplication.jar"
 * Some attrinute of the jar manifest can be use to define synchro behaviours :
 *   - OpenMIS-AppUpdate-Version : define the version of the jar. If the server side version is different from the terminal version
 *      the server version is synchronized on the device. To avoid synchronisatoin after the application installation 
 *      put an empty jar on the device with a manifest with the server version as the server or put no jar on the server side.
 *      After installation, the jar is deleted but the version is stored in the device persistent property : org.openmobileis.synchro.bundle.fullupdate.version
 *   - OpenMIS-PostInstall-Class : a class (implements FullUpdatePostInstallInterface) that is executed after installation before open mobile is Framework is started (webserver).
 *   
 *   After the synchronization if a update file is available, an update program is call to do the update.
 *   The update program is define with the property org.openmobileis.bundle.update.updateprg and the prg parameter
 *   with the property org.openmobileis.bundle.update.updateprgparam
 *   If the properties aren't defined the update is not done.
 *  
 * @author phil
 *
 */
public final class FullAppUpdateListener extends DefaultOpenMSPSyncListener {
	
	private boolean restartAppli = false;
  private boolean syncGoOn = false;
  protected String synchroPath;

	/***
	 * Module that synchronize the application jar.
	 * The jar are copied in the user.dir/temp directory.
	 * The application restarted and the jar unzipped in at the root path defined in the jar META_INF/rootunzippath directive
	 * Use to update full application
	 */
	public FullAppUpdateListener() {
		//update jar info if exist.
		ApplicationUpdateJarUtil util = new ApplicationUpdateJarUtil();
		try	{
			util.readJarMetainf();
			util.updateInstallClass();
			util.deleteUpdateJar();
			ServiceManager.getManager().loadService(new ApplicationUpdateStatusService());
		} catch (IOException ex)	{
			LogManager.traceError(0, "FullAppUpdateListener init error during jar read:"+ex.toString());
		}
		
		
		ApplicationUpdateJarUtil.updateStarted = false;
}

  public void initManager() throws ServiceException	{
  }

	/**
	* Return the name used to identiy the of the listener in the synchro processus.
	* This name correspond to the source in the OpenMS protocol.
	*/
	 public String getSyncName()	{
		 return ApplicationUpdateJarUtil.FULL_APP_UPDATE_NAME;
	 }

	/* (non-Javadoc)
	 * @see org.openmobileis.module.terminal.ModuleManagerListener#getJarInstallPath()
	 */
	public String getJarInstallPath() {
		return System.getProperty("user.dir")+"/temp/application.jar";
	}

	/* (non-Javadoc)
	 * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#startSync(org.openmobileis.synchro.security.auth.Credential, org.openmobileis.synchro.client.SynchroDescriptor)
	 */
	public void startSync(Credential cred, SynchroDescriptor synchrodescriptor) throws OpenMSPException {
		super.startSync(cred, synchrodescriptor);
		this.restartAppli = false;
		this.syncGoOn = true;
    synchroPath = synchrodescriptor.getSynchroConduit().getIntallPath();
    synchroPath = synchroPath+File.separator+ApplicationUpdateJarUtil.APPLICATION_JAR_FILENAME;
	}

	/* (non-Javadoc)
	 * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#sendData(org.openmobileis.synchro.openmsp.protocol.Message)
	 */
	public void sendData(Message message) throws OpenMSPException {
    if (!this.isSyncOK()) {
      return; // return if an error occurs during synchro.
    }
    //send jar last modification date in long.
    if (!syncGoOn) { // synchronize data only for the first call.
    return;
    }
    
    try {
      Command getCommand = new Command(Element.GET, this.getSyncName(), this.getSyncName());
      getCommand.setSourceSessionID(0);
  		ApplicationUpdateJarUtil util = new ApplicationUpdateJarUtil();
      ContainerMessage getContainer = new ContainerMessage(getCommand);
      Item newItem = null;
      newItem = new DataItem(
                  Element.ITEM
                  , "Jar Server file name"
                  , util.getJarVersion()
                  , null
                  , null
              );
      getContainer.add(newItem);
      message.add(getContainer);
      syncGoOn = false;
   } catch (Throwable ex)  {
      this.setSynchroStatus(Status.STATUS_FAILED);
      LogManager.trace(new SynchroException("DefaultDbSyncListener "+this.getSyncName()+" Send data exception ", ex));
    }
	}

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveMapCommand(org.openmobileis.synchro.openmsp.protocol.ContainerMessage)
   */
  public void receiveMapCommand(ContainerMessage mapContainer) throws OpenMSPException {

  }

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveResultCommand(org.openmobileis.synchro.openmsp.protocol.ContainerMessage, org.openmobileis.synchro.openmsp.protocol.ContainerMessage)
   */
  public void receiveResultCommand(ContainerMessage resultContainer, ContainerMessage initialCommand) throws OpenMSPException {
//    Result resultCommand = (Result)resultContainer.getElement(); //result
    try {
      if (resultContainer.hasMoreMessage())  {
        DataItem item = (DataItem) resultContainer.nextMessage().getElement();
        if (item.getMetaInformation().equals("jarfile"))	{
	        FileUtilities.moveFile(this.synchroPath, ApplicationUpdateJarUtil.getTerminalUpdateJarPath());
	        //update jar version
	        ApplicationUpdateJarUtil util = new ApplicationUpdateJarUtil();
	        util.readJarMetainf();
	        this.restartAppli = true;   
        }
      }
      
    } catch (Throwable ex)  {
    	//remove all copied jar file to avoid to remove old one. 
  		File file = new File(this.synchroPath);
  		if (file.exists()) file.delete();
  		file = new File(ApplicationUpdateJarUtil.getTerminalUpdateJarPath());
  		if (file.exists()) file.delete();
  		this.setSynchroStatus(Status.STATUS_FAILED);
      throw new OpenMSPException(ex);
    }
  }

	/* (non-Javadoc)
	 * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#endSync()
	 */
	public void endSync() throws OpenMSPException {
		if (this.restartAppli)	{
			Thread thread = new Thread(new Runnable()	{
				public void run() {
					try	{
						Thread.currentThread().sleep(3000); //way 2 second the synchronisation finised.
						//start update PRG
						String prgname = PropertiesManager.getManager().getProperty("org.openmobileis.bundle.update.updateprg");
						if (prgname != null)	{
							String param = PropertiesManager.getManager().getProperty("org.openmobileis.bundle.update.updateprgparam");
							if (param ==null) param = "";
							LogManager.traceError(0, "FullAppUpdateListener endSync restart appli prgname:"+prgname+" param:"+param);
							SystemAPI.getManager().execProgram(prgname, param);
						}
					} catch (Throwable ex)	{
						LogManager.traceError(0, ex);
					}
				};
			});
			ApplicationUpdateJarUtil.updateStarted = true;
			thread.start();
			this.setSynchroStatus(Status.STATUS_OK);
			OpenMSPSynchroManager.getManager().setGlobalSynchroStatut(Status.STATUS_OK_RESTART_APPLICATION);
//			restart appli. show return page and after stop
		}
	}

  /* (non-Javadoc)
   * @see org.openmobileis.synchro.openmsp.client.OpenMSPSyncListener#receiveStatusCommand(org.openmobileis.synchro.openmsp.protocol.Status, org.openmobileis.synchro.openmsp.protocol.ContainerMessage)
   */
  public void receiveStatusCommand(Status statusCommande, ContainerMessage initialCommand) throws OpenMSPException {
//    int cmlRef = statusCommande.getCmdRef();
    int status = statusCommande.getStatus();
    this.setSynchroStatus(status);
  }

}
