/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  2004 Creation P.Delrieu
 *  2004 Modified by Romain Beaugrand
 * 
 */

package org.openmobileis.common.util;

import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.log.*;

import freemarker.template.instruction.ExitInstruction;

import java.io.File;
import java.util.Properties;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Utility classe to allow to store is the properties file misc.properties
 * service properties to be use later. The storage is persistent.
 * Use this classe when you want to store or get specific service properties.
 * The misc.properties file is created during the first start of the embedded server after installation.
 * Stored properties are mostly runtime properties that are not usefull for the first start.
 * To store service proerties that are needed for the first start use the webserver.properties.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class PersistentPropertiesManager {
	private static PersistentPropertiesManager manager;
	private final static String defaultPropsName = "misc.properties";
	private String propsFile;
	private java.util.Properties props = new Properties();

	public PersistentPropertiesManager(String propertiersFileName) {
		this.initPersistentFilePath(propertiersFileName);
		manager = this;
	}

	private PersistentPropertiesManager() {
		String propsFilePath = PropertiesManager.getManager().getProperty("org.openmobileis.common.persistentpropertiesmanager.propfile");
		if (propsFilePath == null) {
			propsFilePath = System.getProperty("user.dir") + File.separator + "WEB-INF" + File.separator + "conf" + File.separator + defaultPropsName;
		}
		this.initPersistentFilePath(propsFilePath);
		manager = this;
	}

	private void initPersistentFilePath(String propertiersFileName) {
		java.io.FileInputStream input = null;
		propsFile = propertiersFileName;
		try {
			input = new java.io.FileInputStream(propertiersFileName);
			props.load(input);
		} catch (java.io.IOException e) {
			LogManager.traceWarning(LogServices.WEBSERVICE, "ServicePropertiesManager :Can not load " + propsFile + ". Create it.");
			props = new Properties();
			this.store();
		} finally {
			if (input != null) {
				try {
					input.close();
				} catch (Exception e) {
					LogManager.traceWarning(LogServices.WEBSERVICE, "ServicePropertiesManager :Can not close " + propsFile);
				}
			}
		}
	}

	public static PersistentPropertiesManager getManager() {
		if (manager == null) {
			synchronized (PersistentPropertiesManager.class) {
				if (manager == null) {
					manager = new PersistentPropertiesManager();
				}
			}
		}
		return manager;
	}

	/**
	  save the property (key, value) for specified serviceName.
	  key must have less or equals than 15 caractere and value less or equals than 30
	 */
	public void saveProperty(String serviceName, String key, String value) {
		props.put(serviceName + key, value);
		this.store();
	}

	/**
	  return the property with specified key for specified service.
	  null if not found.
	 */
	public String getProperty(String serviceName, String key) {
		return (String) props.getProperty(serviceName + key);
	}

	public Array getPropertyKeysForService(String serviceName) {
		Array ret = new Array(10);
		String key = null;
		for (java.util.Enumeration e = props.keys(); e.hasMoreElements();) {
			key = (String) e.nextElement();
			if (key.startsWith(serviceName)) {
				ret.add(key.substring(serviceName.length(), key.length()));
			}
		}
		return ret;
	}

	public void deleteProperty(String serviceName, String key) {
		props.remove(serviceName + key);
		this.store();
	}

	private void store() {
		java.io.FileOutputStream output = null;
		String key = null;
		String val = null;
		try {
			File file = new File(this.propsFile);
			if (!file.exists()) { //test if directory exist and create it.
				int index = this.propsFile.lastIndexOf(File.separator);
				if (index != -1) {
					String dir = this.propsFile.substring(0, index);
					file = new File(dir);
					if (!file.exists()) {
						file.mkdirs();
					}
				}
			}
			output = new java.io.FileOutputStream(this.propsFile);
			// since the method #store is not available in jdk1.1.6, dot it manually
			for (java.util.Enumeration e = props.keys(); e.hasMoreElements();) {
				key = (String) e.nextElement();
				val = (String) props.get(key);
				output.write((key + "=" + val + "\n").getBytes());
			}
			output.flush();
		} catch (java.io.IOException e) {
			LogManager.traceError(LogServices.WEBSERVICE, "PersistentPropertiesManager Impossible to store property file. Error :" + e.toString());
		} finally {
			if (output != null) {
				try {
					output.close();
				} catch (java.io.IOException e) {
					LogManager.traceError(LogServices.WEBSERVICE, "Impossible to close property file : " + propsFile);
				}
			}
		}
	}

}
