/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2006-2008 Ubikis
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  
 *  Modifications :
 *  Creation P.Delrieu
 * 
 */
package org.openmobileis.common.util.file;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import org.openmobileis.common.util.log.LogManager;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Use to transfert command using file between two programs. The first program 
 * write a file and why the second to answer. The second read the file and answer in a file. 
 * The first read the file and return the content.
 * 
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public final class ReadWriteFileTransfertManager {
	
	class ReadWriteWaiter implements Runnable	{
		private File waitFile;
		private Object notifier;
		private long timeout = 0;
		private boolean hastimeout = false;
		
		ReadWriteWaiter(String waitFileName, Object notifier, long timeout)	{
			this.waitFile = new File(waitFileName);
			this.notifier = notifier;
			this.timeout = timeout;
		}
		public void run()	{
			long time =0;
				try	{
					while (true)	{
					Thread.currentThread().sleep(400);
					if (this.waitFile.exists())	{
						Thread.currentThread().sleep(200); //wait end of the writing;
						 synchronized (this.notifier) {
							 this.notifier.notifyAll();
						 }
						return;
					} else if (this.timeout > 0 && time > this.timeout)	{
						this.hastimeout = true;
						 synchronized (this.notifier) {
							 this.notifier.notifyAll();
						 }
						return;
					}
					time+=400;
					}
				} catch (Throwable ex)	{
					LogManager.traceError(0, ex.toString());
				}
		}
		public boolean isHastimeout() {
			return hastimeout;
		}
	}
	private String baseFilePath = null;
	private long timeout = 0; // no timeout.
	/**
	 * Define the directory where the files are writen.
	 */
	public ReadWriteFileTransfertManager(String directory) {
		this.baseFilePath = directory;
	}
	
	/**
	 * Write the command in the inFileName and write for the outFileName. Return the content of 
	 * outFileName.
	 * If the timeout is define, throw IOException if the wait is greater than the timeout
	 * @param command
	 * @param inFileName
	 * @param outFileName
	 * @param waitfile if waitfile is not null, this file is use to wait for. Otherwise it's the out file.
	 * @return
	 */
	public byte[] sendReadWriteCommand(byte[] command, String inFileName, String outFileName, String waitfile) throws IOException	{
		//write the file.
		String inFile = this.baseFilePath+File.separator+inFileName;
		
//		LogManager.traceDebug(0, "ReadWriteFileTransfertManager write file :"+inFile);
		
		 OutputStream out = new BufferedOutputStream(new FileOutputStream(inFile+".tmp"));
		 try	{
		 out.write(command);
		 } finally	{
			 out.flush();
			 out.close();
		 }
		 File toren = new File(inFile+".tmp");
		 toren.renameTo(new File(inFile));
		 
		 //wait for file
		 Object waiter = new Object();
		 String outFile = this.baseFilePath+File.separator+outFileName;
		 if (waitfile == null) waitfile = outFileName;
		 
	//	 LogManager.traceDebug(0, "ReadWriteFileTransfertManager wait file :"+waitfile);
		 
	   ReadWriteWaiter rw = new ReadWriteWaiter(this.baseFilePath+File.separator+waitfile, waiter, this.timeout);
		 Thread th = new Thread(rw);
		 th.start();
		 synchronized (waiter) {
			 try	{
				 if (this.timeout >0) waiter.wait(timeout*2);
				 else waiter.wait();
			 } catch (InterruptedException ex)	{
				 throw new IOException(ex.toString());
			 }
			 if (rw.isHastimeout())	{
				 throw new IOException("Timeout Exception.");
			 }
			 File tout = new File(outFile);
			 if (!tout.exists())	{
				 throw new IOException("Timeout Exception.");
			 }
			 byte[] retb = FileUtilities.readFile(tout);
			 return retb;
		}
	}

	public long getTimeout() {
		return timeout;
	}

	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}
	
	public static void main(String args[]){
		LogManager.registerLogManager(null);
		String workingdir = System.getProperty("user.dir");
		ReadWriteFileTransfertManager manager = new ReadWriteFileTransfertManager(workingdir);
//		manager.setTimeout(2000);
		try	{
			byte[] testb = manager.sendReadWriteCommand(new String("test file data").getBytes(), "int.test", "out.test", null);
			System.out.println(new String(testb));
		} catch (Throwable ex)	{
			ex.printStackTrace();
		}
	}

}
