/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *
 */
package org.openmobileis.database.fastobjectdb;

import java.util.Hashtable;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.exception.DatabaseException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.db.FODBCollection;
import org.openmobileis.database.fastobjectdb.db.crypto.FODBCypher;

/**
 *
 * Factory pattern to manage one FODb inside an application. Use to get the current db factory
 * instance. Use the PropertiesManager fastobjectdb.database.path and fastobjectdb.database.name properties to define db path and name.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */

public class FastObjectDBManager implements FastObjectDBManagerService {
	protected static Hashtable cypherList;

	protected FastObjectDB db;
	protected boolean fodbStarted = false;
	protected static FastObjectDBManager manager;

	protected FastObjectDBManager() {

	}

	public final static void registerManager(FastObjectDBManager newManager) {
		if (manager == null) {
			manager = newManager;
			ApplicationContextManager.getManager().addManager(manager);
		} else {
			LogManager.traceInfo(0, "SessionContextManager registerManager already done no use");
		}
	}

	public final static FastObjectDBManager getManager() {
		if (manager == null) {
			manager = new FastObjectDBManager();
			ApplicationContextManager.getManager().addManager(manager);
		}
		return manager;
	}

	/**
	 * return the current FODB instance. Open or create it if not done.
	 *
	 * @return the current database instance.
	 * @throws DatabaseException
	 */
	public synchronized FastObjectDB getCurrentFODB() throws DatabaseException {
		if (db == null) {
			String dbPath = PropertiesManager.getManager().getProperty("fastobjectdb.database.path");
			String dbname = PropertiesManager.getManager().getProperty("fastobjectdb.database.name");
			if ((dbPath == null) || (dbname == null)) {
				throw new DatabaseException("FODB ERROR : FODB path or name property is not set. Can't create db");
			}
			synchronized (FastObjectDBManager.class) {
				if (db == null) {
					try {
						db = FastObjectDB.open(dbPath, dbname);
						ApplicationContextManager.getManager().addManager(db);
						fodbStarted = true;
					} catch (Throwable ex) {
						throw new DatabaseException(ex);
					}
				}
			}
		}
		return db;
	}

	public synchronized void CloseCurrentFODB() throws DatabaseException {
		db.getTransactionManager().begin();
		FODBCollection[] colList = db.getDatabaseCollectionArray();
		for (int i = 0; i < colList.length; i++) {
			db.getTransactionManager().enterTransaction(colList[i].getName());
		}
		db = null;
		fodbStarted = false;
	}

	public synchronized boolean isFodbStarted() {
		return fodbStarted;
	}

	/**
	 * Use to reload all db file headers.
	 */
	public synchronized void flushDB() throws DatabaseException {
		if (db != null) {
			db.getTransactionManager().begin();
			FODBCollection[] colList = db.getDatabaseCollectionArray();
			cypherList = new Hashtable();
			for (int i = 0; i < colList.length; i++) {
				db.getTransactionManager().enterTransaction(colList[i].getName());
				FODBCypher cypher = colList[i].getCollectionCypher();
				if (cypher != null) {
					cypherList.put(colList[i].getName(), cypher);
				}
			}
			fodbStarted = false;
			db = null;
		}
		FastObjectDBManager.getManager().getCurrentFODB();
	}

	/**
	 * reset all db collection to empty collection.
	 */
	public synchronized void resetDB() throws DatabaseException {
		if (db != null) {
			db.getTransactionManager().begin();
			FODBCollection[] colList = db.getDatabaseCollectionArray();
			for (int i = 0; i < colList.length; i++) {
				db.getTransactionManager().enterTransaction(colList[i].getName());
				db.removeCollecionFile(colList[i].getName());
			}
			db.getTransactionManager().commit();
		}

	}

}
