/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *
 */

package org.openmobileis.embedded.util;

import org.openmobileis.common.util.exception.OpenMISException;
import org.openmobileis.common.util.log.LogManager;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description:  Helper class to access native OS specific task that is not correcly
 * done by the JVM.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class PPCSystemAPI implements ISystemAPI{

	static	{
		try {
			System.loadLibrary("ppcsystemapi");
		} catch (Throwable exi) {
			exi.printStackTrace();
			LogManager.traceError(0, exi);
		}
	}
	// Define the root register key in long format for the native method
	public static final long HKEY_CLASSES_ROOT = 0x80000000;
	public static final long HKEY_CURRENT_USER = 0x80000001;
	public static final long HKEY_LOCAL_MACHINE = 0x80000002;

  /**
  * Execute a program on mobile and wait until the program ended
  * @param: String path of the program
  * @param: Arguments separated by space
  */
  public native int execProgramAndWait(String path, String args) throws OpenMISException;

  /**
  * Execute a program on mobile.
  * @param: String path of the program
  * @param: Arguments separated by space
  */
  public native int execProgram (String path, String args) throws OpenMISException;
  
  /**
   * Return an unique Id identifying the terminal. 
   * @return
   */
  public native String getSystemUniqueID();
  
  // Get the type of the registry
  private native Long NativeReadPPCRegistryType(long hkey,String regkey,String entity);
  // Return value in string format
  private native String NativeReadPPCRegistryString(long hkey,String regkey,String entity);
  // Return value in a hide "long" format
  private native String NativeReadPPCRegistryLong(long hkey,String regkey,String entity);
  // Return value in byte format
  private native Byte NativeReadPPCRegistryBin(long hkey,String regkey,String entity);
  // Return value in array-string format
  private native String NativeReadPPCRegistryMultiString(long hkey,String regkey,String entity);

  // Method to get value type
  public  Long readPPCRegistryType(long hkey,String regkey,String entity)	{
	  Long nativetype = NativeReadPPCRegistryType(hkey,regkey,entity);
	  return nativetype;
  }
  
  public  String readPPCRegistryMultiString(long hkey,String regkey,String entity)	{
	  String nativetype = NativeReadPPCRegistryMultiString(hkey,regkey,entity);
	  return nativetype;
  }
  
  
  // Method to get value string
  public  String readPPCRegistryString(long hkey,String regkey,String entity)	{
	  String nativetype = NativeReadPPCRegistryString(hkey,regkey,entity);
	  return nativetype;
  }
  
  // Method to get value long
  public  long readPPCRegistryLong(long hkey,String regkey,String entity)	{
	  String nativetype = NativeReadPPCRegistryLong(hkey,regkey,entity);
	  char chartemp1 = nativetype.charAt(0);
	  char chartemp2 = nativetype.charAt(1);
	  int i1 = + chartemp1;
	  int i2 = + chartemp2;
	  long i3 = (i2 & 0xffffffffL)*65536;		// Avoid the signed value
	  long i4 = i1 + i3;
	  return i4;
  }
  
  // Method to get value byte
  public Byte readPPCRegistryBin(long hkey,String regkey,String entity)	{
	  Byte nativetype = NativeReadPPCRegistryBin(hkey,regkey,entity);
	  return nativetype;
  }
   /***
   * Main method to get registry key value
   * Ex : readPPCRegistryValue("HKEY_LOCAL_MACHINE","Time","TimeZoneInformation")
   */
  public String readPPCRegistryValue (String register,String regkey,String entity){
	  String valuestring=null;
	  long valuelong;
	  Byte valuebyte;
	  long hkey = 0;
	  if(register.startsWith("HKEY_CLASSES")){
		  hkey = HKEY_CLASSES_ROOT;
	  }if(register.startsWith("HKEY_LOCAL")){
		  hkey = HKEY_LOCAL_MACHINE;
	  }if(register.startsWith("HKEY_CURRENT")){
		  hkey = HKEY_CURRENT_USER;
	  };
	  Long type = readPPCRegistryType(hkey, regkey, entity);
	  //Depend on the return type of readed regkey
	  //for "String - REG_SZ"
	  if((type.intValue()==1)||(type.intValue()==7)){
		  valuestring = readPPCRegistryString(hkey, regkey, entity);
	  }
	  //for "Long - REG_DWORD"
	  if(type.intValue()==4){
		  valuelong = readPPCRegistryLong(hkey, regkey, entity);
		  valuestring =String.valueOf(valuelong);
	  }
	  //for "Bin - REG_BINARY"
	  if(type.intValue()==3){
		  int temp=0;
		  valuebyte=readPPCRegistryBin(hkey, regkey, entity);
		  valuestring = valuebyte.toString();  
		  StringBuffer buf;
		  buf = new StringBuffer();
		  for (int pos = 0; temp<((valuestring.length()/6)/2); pos+=6) 
		  {			
			temp++;  
		  	String substring = valuestring.substring(pos, pos+2);
			char c = (char) Integer.parseInt(substring, 16);
		 	buf.append(c);
			   
		  }		
		  valuestring = buf.toString();
          }
      return valuestring;
  }
  
  
  /**
  * Get time stamp, same behavior as system.currentTimeMillis in milli
  * Used to fix Creme V3.1 Bug
  */
  public long currentTimeInMillis() {
	  return System.currentTimeMillis();
  }
}