/*
 * OpenMobileIS - a free Java(TM) Framework for mobile applications Java(TM)
 * Copyright (C) 2004-2006 Philippe Delrieu
 * All rights reserved.
 * Contact: pdelrieu@openmobileis.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 *  Author : Philippe Delrieu
 *  Modified by Manuel Gomez 2007.
 *
 */

package org.openmobileis.synchro.openmsp.client;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.zip.GZIPOutputStream;

import org.openmobileis.common.context.ApplicationContextManager;
import org.openmobileis.common.context.Plateform;
import org.openmobileis.common.intl.IntlResourceManager;
import org.openmobileis.common.user.UserNotFoundException;
import org.openmobileis.common.util.PropertiesManager;
import org.openmobileis.common.util.PersistentPropertiesManager;
import org.openmobileis.common.util.codec.GeneralCoder;
import org.openmobileis.common.util.collection.Array;
import org.openmobileis.common.util.exception.SynchroException;
import org.openmobileis.common.util.log.LogManager;
import org.openmobileis.database.fastobjectdb.synchro.client.SynchroListenerAddNotifier;
import org.openmobileis.synchro.client.SynchroDescriptor;
import org.openmobileis.synchro.client.SynchroProcessor;
import org.openmobileis.synchro.openmsp.OpenMSPException;
import org.openmobileis.synchro.openmsp.client.conduit.ConduitParameter;
import org.openmobileis.synchro.openmsp.client.core.ListenerOrdonancer;
import org.openmobileis.synchro.openmsp.client.core.OpenMSPSyncMessageFactory;
import org.openmobileis.synchro.openmsp.protocol.Message;
import org.openmobileis.synchro.openmsp.protocol.MessageFactory;
import org.openmobileis.synchro.openmsp.protocol.Status;
import org.openmobileis.synchro.security.auth.Credential;

/**
 * Title:        OpenMobileIS project source <BR>
 * Description: Default SynchroManager for OpenMSP protocol.
 * @author	Philippe Delrieu
 * @since	JDK 1.1
 * @version 1.0.
 */
public class OpenMSPSynchroManager implements SynchroProcessor {
  public static final String OPENMSPPROCESSORNAME="OpenMSPSync";

  private static OpenMSPSynchroManager manager;
  private ListenerOrdonancer listenerOrdonancer;
  private ListenerOrdonancer moduleOrdonancer;
  private OpenMSPSyncMessageFactory messageFactory;
  private String lastSynchroLogin;
  private int perglobalSynchroStatut;
  private SynchroListenerAddNotifier addlistener;
  private OpenMSPSynchroProcessListener processListener;
  private boolean continuesynchro = false;
  private boolean synchroPending = false;

  /**
   *
   */
  private OpenMSPSynchroManager() {
    super();
    this.listenerOrdonancer = new ListenerOrdonancer();
    this.moduleOrdonancer = new ListenerOrdonancer();
    this.messageFactory = new OpenMSPSyncMessageFactory();
    String ss = PersistentPropertiesManager.getManager().getProperty(OPENMSPPROCESSORNAME, "globalSynchroStatut");
    if (ss!=null){
    	this.perglobalSynchroStatut =  Integer.parseInt(ss);
    }
  }

  public static OpenMSPSynchroManager getManager() {
    if (manager == null)  {
      synchronized (OpenMSPSynchroManager.class) {
        if (manager == null)  {
          manager = new OpenMSPSynchroManager();
          ApplicationContextManager.getManager().addManager(manager);
        }
      }
    }
    return manager;
  }
  
  public void setOpenMSPSynchroProcessListener(OpenMSPSynchroProcessListener listener)	{
  	this.processListener = listener;
  }

  public void clearListeners(){
	  this.listenerOrdonancer.clearListeners();
  }

  public String getProcessorName()  {
    return OpenMSPSynchroManager.OPENMSPPROCESSORNAME;
  }

  public void notifyFailure(String listernerName) {
    OpenMSPSyncListener listener = this.getListenerByName(listernerName);
    if (listener != null) {
      listener.notifySynchroFailure();
      listenerOrdonancer.notifySynchroError(listener.getSyncName());
      moduleOrdonancer.notifySynchroError(listener.getSyncName());
    }
    this.setGlobalSynchroStatut(Status.STATUS_FAILED);
  }
  
  /**
   * Return an Array of all the listeners contains
   * @return array of listeners
   * Added by Manuel GOMEZ
   */
  public Array getAllListeners() {
	  return listenerOrdonancer.getAllListeners();
  }

  /**
   * Return the registered listener with specified name.
   * @param listenerName
   * @return registered listener with specified name.
   */
  public OpenMSPSyncListener getListenerByName(String listenerName) {
    return listenerOrdonancer.getListenerByName(listenerName);
  }
  
  /**
   * Remove the listener with the specified name. It remove all dependent listener too.
   * Return a boolean to inform the succed or not of the remove operation
   * @param listenerName
   * @return succed or not boolean
   */
  public boolean removeListenerByName(String listenerName) {
	  return listenerOrdonancer.removeListenerByName(listenerName);
  }
  
  /**
   * Inactivate the specified listener name. Depend listener are not Inactivate
   * @param listenerName
   * @return succed or not boolean
   */
  public void inactivateListenerByName(String listenerName) {
	  this.listenerOrdonancer.setInactivateListenerByName(listenerName, true);
  }
  
  /**
   * Activate the specified listener name. Depend listener are not Activate if previously inactivate
   * @param listenerName
   * @return succed or not boolean
   */
  public void activateListenerByName(String listenerName) {
	  this.listenerOrdonancer.setInactivateListenerByName(listenerName, false);
  }

  /**
   * Return the registered module listener with specified name.
   * Module listener participate to module synchro
   * Module synchro is the synchro of module specific files to manage module update.
   *
   * @param listenerName
   * @return
   */
  public OpenMSPSyncListener getModuleListenerByName(String listenerName) {
    return moduleOrdonancer.getListenerByName(listenerName);
  }
  
  /**
   * restart the synchro when the current synchro finish. Only effectiveif call during a synchronization.
   */
  public synchronized void restartSynchro()	{
  	this.continuesynchro = true;
    LogManager.traceDebug(0, "OpenMSPSynchroManager RESTARTSYNCHRO");
  }
  
  /***
   * This method block until the synchro end. Return immediatly if the synchro is not started.
   */
  public synchronized void waitSynchroEnd()	{
  	if (synchroPending)	{
	    LogManager.traceDebug(0, "OpenMSPSynchroManager waitSynchroEnd BEFORE WAIT");
	    try {
	    	this.wait();
	    } catch(Throwable ex){
		    LogManager.traceDebug(0, "OpenMSPSynchroManager waitSynchroEnd ERROR");
	    	LogManager.traceDebug(0, ex.toString());
	    }
	    LogManager.traceDebug(0, "OpenMSPSynchroManager waitSynchroEnd END WAIT");
  	}
  }

  public void doSynchro(Credential cred, SynchroDescriptor descriptor) throws SynchroException {
  	synchronized (this)	{
	    LogManager.traceDebug(0, "OpenMSPSynchroManager doSynchro START");
  		synchroPending = true;
  	}
  	try	{
  	this.setGlobalSynchroStatut(Status.STATUS_SYNC_IN_PROGRESS);
    if (processListener != null) processListener.notifySynchroStart();
    this.lastSynchroLogin = cred.getPrincipal();

    //save login
    PersistentPropertiesManager.getManager().saveProperty("OpenMSPSynchroManager", "login", this.lastSynchroLogin);
    
    this.continuesynchro = true;
    while (this.continuesynchro)	{
	    synchronized(this)	{
	    	this.continuesynchro = false;
	    }
	    ListenerOrdonancer currentOrdonnancer = this.listenerOrdonancer;
	    currentOrdonnancer.beginSynchro();
	    this.messageFactory.beginSynchro(cred, descriptor);
	    Array listenerList = currentOrdonnancer.getNextListenerList();
//	    LogManager.traceDebug(0, "OpenMSPSynchroManager synchro listener list size :"+listenerList.size()+" continuesynchro:"+this.continuesynchro);
	    boolean moduleSynchro = false;
	    if (listenerList.size() == 0)   {
	      currentOrdonnancer.endSynchro();
	      currentOrdonnancer = this.moduleOrdonancer;
	      currentOrdonnancer.beginSynchro();
	      listenerList = currentOrdonnancer.getNextListenerList();
	      moduleSynchro = true;
	    }
	    try {
	      while (listenerList.size() != 0)  {
	        int listsize = listenerList.size();
	  	    LogManager.traceDebug(0, "OpenMSPSynchroManager synchro listener list size in while :"+listsize);
	        for (int i=0; i<listsize; i++)  {
	          OpenMSPSyncListener listener = (OpenMSPSyncListener) listenerList.get(i);
	          if (processListener != null) processListener.notifyListenerStartSync(listener.getSyncName());
	        //  LogManager.traceDebug(0, "OpenMSPSynchroManager start sync for listener  :"+listener.getSyncName());
	          listener.startSync(cred, descriptor);
	        }
	        while (true) {
	         if (processListener != null) processListener.notifyGenerateOpenMSPMessage();
	          Message message = this.messageFactory.getOpenMSPMessage(listenerList);
	          if (message.isEmpty()) {
	            //get next listener phase.
	            break;
	          }
	          String xmlString = message.encode();
	          ConduitParameter[] parameters = new ConduitParameter[] {new ConduitParameter("openMISGroup", descriptor.getSynchroGroup())};
	          if (processListener != null) processListener.notifyBeforeSendMessage();
	          String reqResult = descriptor.getSynchroConduit().sendRequest(parameters, new String(compressAndEncodeRequest(xmlString)), descriptor.getServerURL());
	          // third step : Process responses
	          if (processListener != null) processListener.notifyReceiveMessageReturn();
	          Message receivedMessage = MessageFactory.getFactory().getMessage(reqResult);
	          String target = receivedMessage.getHeader().getTarget();
	          if (processListener != null) processListener.notifyBeforeProcessMessageReturn();
	          if (!target.equals(OpenMSPSyncMessageFactory.OpenMSPSOURCE)) {
	            throw new OpenMSPException("Bad OpenMSP Sync Source in Server answer message. Synchro Abort.");
	          } else {
	            this.messageFactory.receiveOpenMSPMessage(listenerList, receivedMessage);
	          }
	          if (processListener != null) processListener.notifyAfterProcessMessageReturn();
	        }
	        for (int i=0; i<listsize; i++)  {
	          OpenMSPSyncListener listener = (OpenMSPSyncListener) listenerList.get(i);
	          if (processListener != null) processListener.notifyListenerEndSync(listener.getSyncName());
	          listener.endSync();
	        }
	        listenerList = currentOrdonnancer.getNextListenerList();
	        if (listenerList.size() == 0 && !moduleSynchro)   {
	          currentOrdonnancer.endSynchro();
	          currentOrdonnancer = this.moduleOrdonancer;
	          currentOrdonnancer.beginSynchro();
	          listenerList = currentOrdonnancer.getNextListenerList();
	          moduleSynchro = true;
	        }
	      }
	      if (processListener != null) processListener.notifyEndProcessListener();
	      if (this.getGlobalSynchroStatut() == Status.STATUS_SYNC_IN_PROGRESS) this.setGlobalSynchroStatut(Status.STATUS_OK);
	    } catch (UserNotFoundException ex)  {
	      notifyFailureToListeners(listenerList);
	      this.setGlobalSynchroStatut(Status.STATUS_UNAUTHORIZED);
	      if (processListener != null) processListener.notifySynchroFailure(Status.STATUS_UNAUTHORIZED);
	      throw new OpenMSPException(ex);
	    } catch (Throwable ex)  {
	      notifyFailureToListeners(listenerList);
	      if (processListener != null) processListener.notifySynchroFailure(Status.STATUS_FAILED);
	      throw new OpenMSPException(ex);
	    } finally {
	      this.listenerOrdonancer.endSynchro();
	      this.moduleOrdonancer.endSynchro();
	      this.messageFactory.endSynchro();
	      if (processListener != null) processListener.notifySynchroEnd();
	    }
    }
  	} finally	{
  		synchronized(this){
  			this.synchroPending = false;
  			this.notifyAll();
  	    LogManager.traceDebug(0, "OpenMSPSynchroManager doSynchro END");
  		}
  	}
  }

  /**
   * register a listener that will be notified when a synchrolistener is added.
   */
  public void registerSynchroListenerAddNotifier(SynchroListenerAddNotifier listener) {
    addlistener = listener;
  }

  /**
   * register a new listener and its depends that will participate to the synchro.
   * Listener's depends are an array of listener name that must be synchronized before
   * the new listener.
   * There is no specific order to register the listener and its dependance.
   * If the SynchroListenerAddNotifier is registered the listener is registered only if the isBlocking() call
   * return true. In other case nothing is done.
   * The listener initListener() method is call before in all case.
   * @param listener : listener to register
   * @param depends : list of listener name
   */
  public void addListener(OpenMSPSyncListener listener, String[] depends) {
   // LogManager.traceDebug(0, "OpenMSPSynchroManager add listener  :"+listener.getSyncName());
  	// init listener.
  	try	{
  		listener.initListener();
  	}catch	(OpenMSPException ex)	{
  		LogManager.traceError(0, ex);
  		return;
  	}

    if (addlistener != null) {
      addlistener.notifyAddListener(listener, depends);
      if (!addlistener.isBlocking())	{
      	listenerOrdonancer.addListener(listener, depends);
      }
    } else  {
      listenerOrdonancer.addListener(listener, depends);
    }
  }



  /**
   * register a new listener and its depends that will participate to the module synchro.
   * Module synchro is the synchro of module specific files to manage module update.
   * Listener's depends are an array of listener name that must be synchronized before
   * the new listener.
   * Thier is no specific order to register the listener and its dependance.
   * @param listener : listener to register
   * @param depends : list of listener name
   */
  public void addModuleListener(OpenMSPSyncListener listener, String[] depends) {
    moduleOrdonancer.addListener(listener, depends);
  }

  private void notifyFailureToListeners(Array listenerList) {
    for (int i=0; i<listenerList.size(); i++) {
      OpenMSPSyncListener listener = (OpenMSPSyncListener)listenerList.get(i);
      listener.notifySynchroFailure();
      listenerOrdonancer.notifySynchroError(listener.getSyncName());
      moduleOrdonancer.notifySynchroError(listener.getSyncName());
    }
    this.setGlobalSynchroStatut(Status.STATUS_FAILED);
  }

  private byte[] compressAndEncodeRequest(String request) throws IOException {
    // compresse data
    byte[] compressedData = null;
    ByteArrayOutputStream byteStream = new ByteArrayOutputStream(512);
    GZIPOutputStream zipStream = new GZIPOutputStream(byteStream);
    try {
      zipStream.write(request.getBytes());
    } finally  {
      zipStream.close();
    }
    compressedData = byteStream.toByteArray();

    // encode compressed data
    byte[] returnData = GeneralCoder.encodeBase64(compressedData);
    return returnData;
  }

  public String getLastSynchroLogin() {
    if (this.lastSynchroLogin == null){
      this.lastSynchroLogin = PersistentPropertiesManager.getManager().getProperty("OpenMSPSynchroManager", "login");

    }
    return this.lastSynchroLogin;
  }

  /**
   * return the user agent to use with synchronisation.
   * The user agent can be defined with the property org.openmobileis.synchro.direct.useragent.
   * or if not set, it's constructed with plateform information.
   * default user agent : OpenMSPPlug-<OpenMobileISMajorVersion>.<OpenMobileISMinorVersion>-<OS>[<OSVersion>]-<user language>-<synchro type if present>
   * The synchro type is defined with the property OpenMSPsynchrotype of the descriptor.
   * @return
   */
  public String getSynchroUserAgent(SynchroDescriptor descriptor) {
    String usera = PropertiesManager.getManager().getProperty("org.openmobileis.synchro.direct.useragent");
    if (usera == null)  {
      Plateform plateform = ApplicationContextManager.getManager().getApplicationContext().getPlateform();
      StringBuffer useragent = new StringBuffer("OpenMSPPlug-");
      useragent.append(plateform.getOpenMobileISMajorVersion()).append('.').append(plateform.getOpenMobileISMinorVersion()).append('-');
      useragent.append(plateform.getOS()).append('[').append(plateform.getOSVersion()).append("]-");
      String language = IntlResourceManager.getManager().getManagerLocaleString();
      useragent.append(language);
      String synctype = descriptor.getProperty("OpenMSPsynchrotype");
      if (synctype != null) {
        useragent.append('-').append(synctype);
      }
      usera = useragent.toString();
    }
    return usera;
  }

  public int getGlobalSynchroStatut() {
    return this.perglobalSynchroStatut;
  }

  public void setGlobalSynchroStatut(int status) {
  	PersistentPropertiesManager.getManager().saveProperty(OPENMSPPROCESSORNAME, "globalSynchroStatut", Integer.toString(status));
    perglobalSynchroStatut = status;
  }

}
